use anyhow::{Context, Result};
use serde::{Deserialize, Serialize};
use serde_json::Value;
use std::collections::{BTreeMap, HashSet};
use std::fs;
use std::path::{Path, PathBuf};

use crate::downloader; // Import the downloader module to use its helpers

/// Represents a single rule, capturing the URL, tag, and any other fields.
#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct Rule {
    pub url: String,
    pub tag: String,
    // Capture any other fields in the rule object
    #[serde(flatten)]
    pub other_fields: BTreeMap<String, Value>,
}

/// Represents the "route" object in the config.
#[derive(Serialize, Deserialize, Debug)]
struct Route {
    #[serde(default, rename = "rule_set")]
    rule_set: Vec<Rule>,
}

/// Represents the top-level structure of the configuration file.
#[derive(Serialize, Deserialize, Debug)]
pub struct Config {
    route: Route,
    // Capture any other top-level fields
    #[serde(flatten)]
    other_fields: BTreeMap<String, Value>,
}

impl Config {
    /// Loads and parses the config file from a given path.
    pub fn load(path: &Path) -> Result<Self> {
        let content = fs::read_to_string(path)
            .with_context(|| format!("Failed to read config file from '{}'", path.display()))?;
        let config: Config = serde_json::from_str(&content)
            .context("Failed to parse JSON. Please check the config file structure.")?;
        Ok(config)
    }

    /// Extracts all URLs from the rule_set.
    pub fn extract_urls(&self) -> Vec<String> {
        self.route
            .rule_set
            .iter()
            .map(|rule| rule.url.clone())
            .collect()
    }

    /// Generates a set of the absolute paths of all files that are expected
    /// to exist in the output directory based on the config.
    pub fn get_expected_files(&self, output_dir: &Path) -> Result<HashSet<PathBuf>> {
        self.route
            .rule_set
            .iter()
            .map(|rule| {
                let filename = downloader::url_to_filename(&rule.url)?;
                Ok(output_dir.join(filename))
            })
            .collect()
    }
}
