use anyhow::{Context, Result};
use futures::{StreamExt, stream};
use std::collections::HashSet;
use std::fs;
use std::path::{Path, PathBuf};
use tracing::{debug, error, info, instrument, warn};

/// A report summarizing the outcome of the download process.
#[derive(Debug)]
pub struct DownloadReport {
    pub successful: usize,
    pub failed: usize,
}

/// Extracts a filename from a URL string.
pub fn url_to_filename(url: &str) -> Result<String> {
    url.split('/')
        .last()
        .filter(|s| !s.is_empty())
        .map(String::from)
        .with_context(|| format!("Could not determine filename for URL '{}'", url))
}

/// Scans the output directory and removes any files not in the expected set.
#[instrument(skip(output_dir, expected_files))]
pub fn cleanup_stale_files(output_dir: &Path, expected_files: &HashSet<PathBuf>) -> Result<()> {
    info!("Checking for stale rule files...");
    let mut cleaned_count = 0;

    for entry in fs::read_dir(output_dir)
        .with_context(|| format!("Failed to read output directory '{}'", output_dir.display()))?
    {
        let entry = entry?;
        let path = entry.path();

        if !path.is_file() {
            continue;
        }

        if !expected_files.contains(&path) {
            match fs::remove_file(&path) {
                Ok(_) => {
                    info!(path = %path.display(), "Removed stale file");
                    cleaned_count += 1;
                }
                Err(e) => {
                    warn!(path = %path.display(), error = %e, "Failed to remove stale file");
                }
            }
        }
    }

    if cleaned_count > 0 {
        info!(count = cleaned_count, "Cleanup complete.");
    } else {
        debug!("No stale files found.");
    }

    Ok(())
}

/// Downloads a list of URLs concurrently, with a specified limit.
#[instrument(skip_all, fields(count = urls.len(), concurrency))]
pub async fn download_all_rules(
    urls: &[String],
    output_dir: &Path,
    concurrency: usize,
) -> Result<DownloadReport> {
    info!("Starting download process...");
    let client = reqwest::Client::new();
    let mut successful = 0;
    let mut failed = 0;

    let fetches = stream::iter(urls)
        .map(|url| {
            let client = client.clone();
            let output_path = output_dir.to_path_buf();
            async move {
                match download_rule(&client, url, &output_path).await {
                    Ok(path) => {
                        debug!(path = %path.display(), "Rule saved successfully");
                        Ok(())
                    }
                    Err(e) => {
                        error!(url, error = ?e, "Download failed");
                        Err(())
                    }
                }
            }
        })
        .buffer_unordered(concurrency)
        .collect::<Vec<_>>()
        .await;

    for result in fetches {
        if result.is_ok() {
            successful += 1;
        } else {
            failed += 1;
        }
    }

    Ok(DownloadReport { successful, failed })
}

/// Downloads a single file from a URL to a destination directory.
#[instrument(skip(client, output_dir), fields(url))]
async fn download_rule(client: &reqwest::Client, url: &str, output_dir: &Path) -> Result<PathBuf> {
    let file_name = url_to_filename(url)?;
    debug!(file_name, "Downloading rule");

    let final_path = output_dir.join(&file_name);
    let tmp_path = output_dir.join(format!("{}.tmp", file_name));

    let response = client.get(url).send().await?.error_for_status()?;
    let content = response.bytes().await?;

    fs::write(&tmp_path, &content)
        .with_context(|| format!("Failed to write temporary file {}", tmp_path.display()))?;

    fs::rename(&tmp_path, &final_path)
        .with_context(|| format!("Failed to rename temp file to {}", final_path.display()))?;

    Ok(final_path)
}
